MODULE UnitConversions;

 (*
    UnitConversions - Perform metric/imperial conversions.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT el := UnixExecLists, f := Files, x := ExComplexLib0, tr := TextRider,
       IntStr, xn := ExNumbers, Str := Strings, y := SYSTEM;


CONST
  Length* = 0; Area* = 1; Volume* = 2; Mass* = 3; Pressure* = 4;
  Speed* = 5; Time* = 6; Mileage* = 7; Temperature* = 8;

  Celsius = 0; Fahrenheit = 1; Kelvin = 2; Rankin = 3;

TYPE
  String = ARRAY 80 OF CHAR;
  ConvBaseRec = RECORD
    Units: String;
    Scale: x.ExComplexType
  END;

  (* Note: the size of this array is fixed because the compiler
     has a bug where it generates range errors (on valid ranges)
     when indexing a dynamic array of this type. *)
  ConvBaseArray=ARRAY 32 OF ConvBaseRec;
  ConvBasePtr=POINTER TO ConvBaseArray;
  BaseRec = RECORD
    size: INTEGER;
    k: ConvBasePtr
  END;

VAR
  Category-: SHORTINT;
  From-, To-: INTEGER;
  ConvBase: ARRAY Temperature+1 OF BaseRec;
  ConvList: el.List;
  fscale,kscale: x.ExComplexType;


PROCEDURE Convert* (VAR n: x.ExComplexType);
(* Convert n according to the current "From" type to
   the current "To" type. *)
BEGIN
  IF From=To THEN RETURN END;   (* no conversion needed *)
  IF Category=Temperature THEN
    IF From=Fahrenheit THEN x.ExSub(n, n, fscale)
    ELSIF From=Kelvin  THEN x.ExSub(n, n, kscale)
    END
  END;
  x.ExMult(n, n, ConvBase[Category].k[To].Scale);
  x.ExDiv(n, n, ConvBase[Category].k[From].Scale);
  IF Category=Temperature THEN
    IF To=Fahrenheit THEN x.ExAdd(n, n, fscale)
    ELSIF To=Kelvin  THEN x.ExAdd(n, n, kscale)
    END
  END;
END Convert;


PROCEDURE SetCategory * (cat: SHORTINT);
BEGIN
  IF (cat>=0) & (cat<=Temperature) THEN Category:=cat END
END SetCategory;

PROCEDURE CreateList * (): LONGINT;
VAR cnt: LONGINT;
BEGIN
  el.Dispose(ConvList);
  FOR cnt:=0 TO ConvBase[Category].size-1 DO
    el.Append(ConvList, ConvBase[Category].k[cnt].Units)
  END;
  RETURN y.ADR(ConvList)
END CreateList;

PROCEDURE SetFrom * (fromType: INTEGER);
BEGIN
  IF (fromType>=0) & (fromType<ConvBase[Category].size) THEN
    From:=fromType
  END
END SetFrom;

PROCEDURE SetTo * (toType: INTEGER);
BEGIN
  IF (toType>=0) & (toType<ConvBase[Category].size) THEN
    To:=toType
  END
END SetTo;


PROCEDURE SetScale(cat, ind: LONGINT; s: x.ExComplexType);
BEGIN
  ConvBase[cat].k[ind].Scale:=s
END SetScale;


PROCEDURE SetUnits(cat, ind: LONGINT; s: ARRAY OF CHAR);
BEGIN
  COPY(s, ConvBase[cat].k[ind].Units)
END SetUnits;


PROCEDURE Occurs (s, pat: ARRAY OF CHAR) : INTEGER;
VAR found: BOOLEAN; pos: INTEGER;
BEGIN
  Str.FindNext(pat, s, 0, found, pos);
  RETURN pos
END Occurs;


PROCEDURE AssignUnits(l: ARRAY OF CHAR);
VAR
  str: ARRAY 80 OF CHAR;
  cat, ind: LONGINT; 
  pos: INTEGER;
  res: SHORTINT;
  n: xn.ExNumType;
  cn: x.ExComplexType;
BEGIN
  (* extract the category *)
  IF (l[0]#0X) & (Occurs(l, "#")<0) THEN
    pos:=Occurs(l, ":"); Str.Extract(l, 0, pos, str);
    IntStr.StrToInt(str, cat, res);
    IF res#IntStr.strAllRight THEN cat:=0 END;
    Str.Delete(l, 0, pos+1); IF cat>Temperature THEN cat:=0 END;

    (* extract the index *)
    pos:=Occurs(l, ":"); Str.Extract(l, 0, pos, str);
    IntStr.StrToInt(str, ind, res);
    IF res#IntStr.strAllRight THEN ind:=0 END;
    Str.Delete(l, 0, pos+1);

    (* create an array of the maximum index length *)
    (* NOTE: first index must be the largest *)
    IF ConvBase[cat].k=NIL THEN
      NEW(ConvBase[cat].k); ConvBase[cat].size:=SHORT(ind+1)
    END;

    (* extract the scaling number *)
    pos:=Occurs(l, ":"); Str.Extract(l, 0, pos, str);
    xn.StrToExNum(str, n); x.Init(cn, n, xn.Ex0);
    SetScale(cat, ind, cn); Str.Delete(l, 0, pos+1);

    (* last part of string are the units *)
    SetUnits(cat, ind, l)
  END
END AssignUnits;


PROCEDURE Init*;
VAR
  file: f.File;
  r: tr.Reader;
  cline: ARRAY 132 OF CHAR;
  cnt: LONGINT; 
  res: INTEGER;
  t: xn.ExNumType;
BEGIN
  (* temperature conversions are built in *)
  xn.ExNumb(32, 0, 0, t); x.Init(fscale, t, xn.Ex0);
  xn.ExNumb(273, 0, 0, t); x.Init(kscale, t, xn.Ex0);

  (* initialize the default units *)
  Category:=Temperature; From:=Fahrenheit; To:=Celsius;

  (* clear out the conversion indices *)
  FOR cnt:=0 TO Temperature DO ConvBase[cnt].k:=NIL END;

  (* read external conversion factors *)
  file:=f.Old("Conversion.dat", {f.read}, res);
  IF file#NIL THEN
    r:=tr.ConnectReader(file);
    WHILE (r#NIL) & (r.Res()=tr.done) DO
      r.ReadLine(cline);
      AssignUnits(cline)
    END;
    file.Close
  END
END Init;


BEGIN
  el.Init(ConvList); Init
END UnitConversions.
